//
//
//	HillTerrain.cpp
//
//
//	a class for a heightmap-based terrain generated using the hill algorithm.
//
//	change log:
//	--------
//	10/10/01	bob	first completion
//
//	----------------------------------------------------------------------------
//	copyright (c) 2001 bob nystrom                  [see license at end of file]

#include "HillTerrain.h"

#include <math.h>
#include <assert.h>
#include <new.h>

//	HillTerrain
//	----------------------------------------------------------------------------
CHillTerrain::CHillTerrain(	unsigned short	uSize,
							float			fHillMin,
							float			fHillMax,
							float			fMaxHillHeight,
							unsigned short	uNumHills,
							unsigned short	uFlattening,
							bool			bIsland,
							unsigned int	uSeed )
:	m_uSize( uSize ),
	m_fHillMin( fHillMin ),
	m_fHillMax( fHillMax ),
	m_fMaxHillHeight( fMaxHillHeight ),
	m_uNumHills( uNumHills ),
	m_uFlattening( uFlattening ),
	m_bIsland( bIsland ),
	m_uSeed( uSeed )
{
	// check the parameters
	assert( uSize > 0 );
	assert( fHillMin >= 0.0 );
	assert( fHillMin < fHillMax );
	assert( fHillMax >= 0.0 );
	assert( uFlattening > 0 );
	
	#ifdef _WIN32
		// hack since windows fails to throw bad_alloc on a failed new like
		// it should.
		
		// allocate the map
		m_pMap = new float[ m_uSize * m_uSize ];
		m_pNormals = new C3dVector[ m_uSize * m_uSize ];
		
		if( m_pMap == NULL || m_pNormals == NULL )
		{
			throw( ErrorOutOfMemory() );
		}
		
	#else
		// this is the c++ standard way of behaving
		try
		{
			// allocate the map
			m_pMap = new float[ m_uSize * m_uSize ];
			m_pNormals = new C3dVector[ m_uSize * m_uSize ];
		}
		catch( std::bad_alloc )
		{
			throw( ErrorOutOfMemory() );
		}
	#endif
}


//	~HillTerrain
//	----------------------------------------------------------------------------
CHillTerrain::~CHillTerrain()
{
	delete [] m_pMap;		// free up the cell buffer
	delete [] m_pNormals;	// free up the normals buffer
}


//	Generate
//	----------------------------------------------------------------------------
void
CHillTerrain::Generate( void )
{
	// make sure there is a terrain
	assert( m_pMap != NULL );

	Clear();
	
	// set the seed
	srand( m_uSeed );
	
	// add as many hills as needed
	for( int i = 0; i < m_uNumHills; ++i )
	{
		AddHill();
	}
	
	// now clean it up
	Normalize();
	Flatten();

	// finally generate normals
	GenerateNormals();
}


//	Generate Normals
//	----------------------------------------------------------------------------
void
CHillTerrain::GenerateNormals( void )
{
	// make sure there is a vertex buffer
	assert( m_pNormals != NULL );

	// ______________________________
	// generate normals for all faces
	C3dVector *Afaces = new C3dVector[(m_uSize-1)*(m_uSize-1)];
	C3dVector *Bfaces = new C3dVector[(m_uSize-1)*(m_uSize-1)];

	int x, z, face_pos;
	C3dVector v1, v2;
	
	for(z = 0; z < m_uSize-1; z++)
	{
		for(x = 0; x < m_uSize-1; x++)
		{
			face_pos = x + ( z * (m_uSize-1) );

			// generate normal for first face
			v1.setXYZ( 1, (GetCell(x+1, z+1)-GetCell(x, z)), -1 );
			v2.setXYZ( 0, (GetCell(x, z+1)-GetCell(x, z)), -1 );
			Afaces[face_pos] = v1*v2;
			Afaces[face_pos].normalize();

			// generate normal for second face
			v2.setXYZ( 1, (GetCell(x+1, z)-GetCell(x, z)), 0 );
			Bfaces[face_pos] = v2*v1;
			Bfaces[face_pos].normalize();
		}
	}

	// __________________________________________
	// average face normals to get vertex normals
	for(z = 0; z < m_uSize; z++)
	{
		for(x = 0; x < m_uSize; x++)
		{
			v1.setXYZ(0, 0, 0);					// hold vertex-normal
			face_pos = x + ( z * (m_uSize-1) );

			// Add potential 6 faces
			if (x < m_uSize-1 && z < m_uSize-1) v1 += Afaces[face_pos] + Bfaces[face_pos];
			if (x > 0 && z < m_uSize-1 ) v1 += Bfaces[face_pos-1];
			face_pos = (x-1) + ( (z-1) * (m_uSize-1) );
			if (x < m_uSize-1 && z > 0) v1 += Afaces[face_pos+1];
			if (x > 0 && z > 0) v1 += Afaces[face_pos] + Bfaces[face_pos];

			// add vertex-normal to buffer (it'll get automatically normalized)
			SetNormal(x, z, v1);
		}
	}
	delete[] Afaces;
	delete[] Bfaces;
}


//	Clear
//	----------------------------------------------------------------------------
void
CHillTerrain::Clear( void )
{
	// make sure there is a terrain
	assert( m_pMap != NULL );

	for( int x = 0; x < m_uSize; ++x )
	{
		for( int z = 0; z < m_uSize; ++z )
		{
			SetCell( x, z, 0 );
		}
	}

	C3dVector nothing(0, 0, 0);
	for( int x = 0; x < m_uSize; ++x )
	{
		for( int z = 0; z < m_uSize; ++z )
		{
			SetNormal( x, z, nothing );
		}
	}
}


//	AddHill
//	----------------------------------------------------------------------------
void
CHillTerrain::AddHill( void )
{
	// make sure there is a terrain
	assert( m_pMap != NULL );

	// pick a size for the hill
	float fRadius = RandomRange( m_fHillMin, m_fHillMax );
	
	// pick a centerpoint for the hill
	float x, z;
	if( m_bIsland )
	{
		// island code:
		
		float fTheta = RandomRange( 0, 6.28 );
		// this determines in which direction from the center of the map the
		// hill will be placed.
		
		float fDistance = RandomRange( fRadius/2, m_uSize/2 - fRadius );
		// this is how far from the center of the map the hill be placed. note
		// that the radius of the hill is subtracted from the range to prevent
		// any part of a hill from reaching the very edge of the map.
		
		x = m_uSize/2.0 + cos( fTheta ) * fDistance;
		z = m_uSize/2.0 + sin( fTheta ) * fDistance;
		// converts theta and a distance into x and y coordinates.
	}
	else
	{
		// non-island code:
		
		x = RandomRange( fRadius, m_uSize - fRadius - 1 );
		z = RandomRange( fRadius, m_uSize - fRadius - 1 );
		// Changed by Arsham... This makes sure that the entire hill is on the map
		// and that the edges are all zero height!
	}

	// square the hill radius so we don't have to square root the distance 
	float fRadiusSq = fRadius * fRadius;
	float fDistSq;
	float fHeight;
	
	// find the range of cells affected by this hill
	int xMin = x - fRadius - 1;
	int xMax = x + fRadius + 1;
	// don't affect cell outside of bounds
	if( xMin < 0 )
		xMin = 0;
	if( xMax >= m_uSize )
		xMax = m_uSize - 1;
	
	int zMin = z - fRadius - 1;
	int zMax = z + fRadius + 1;
	// don't affect cell outside of bounds
	if( zMin < 0 ) zMin = 0;
	if( zMax >= m_uSize ) zMax = m_uSize - 1;
	
	// for each affected cell, determine the height of the hill at that point
	// and add it to that cell
	for( int h = xMin; h <= xMax; ++h )
	{
		for( int v = zMin; v <= zMax; ++v )
		{
			// determine how far from the center of the hill this point is
			fDistSq = ( x - h ) * ( x - h ) + ( z - v ) * ( z - v );
			// determine the height of the hill at this point
			fHeight = fRadiusSq - fDistSq;
			
			// don't add negative hill values (i.e. outside the hill's radius)
			if( fHeight > 0 )
			{
				// add the height of this hill to the cell
				OffsetCell( h, v, fHeight );
			}	
		}
	}
}


//	Normalize
//	----------------------------------------------------------------------------
void
CHillTerrain::Normalize( void )
{
	// make sure there is a terrain
	assert( m_pMap != NULL );
	
	float fMin = GetUnscaledCell( 0, 0 );
	float fMax = GetUnscaledCell( 0, 0 );
	
	// find the min and max
	for( int x = 0; x < m_uSize; ++x )
	{
		for( int z = 0; z < m_uSize; ++z )
		{
			float y = GetUnscaledCell( x, z );
			if( y < fMin ) fMin = y;
			if( y > fMax ) fMax = y;
		}
	}

	// avoiding divide by zero (unlikely with floats, but just in case)
	if( fMax != fMin )
	{
		// divide every height by the maximum to normalize to ( 0.0, 1.0 )
		for( int x = 0; x < m_uSize; ++x )
		{
			for( int z = 0; z < m_uSize; ++z )
			{
				SetCell( x, z, ( GetUnscaledCell( x, z ) - fMin ) / ( fMax - fMin ) );
			}
		}
	}
	else
	{
		// if the min and max are the same, then the terrain has no height, so just clear it
		// to 0.0.
		Clear();
	}
}	


//	Flatten
//	----------------------------------------------------------------------------
void
CHillTerrain::Flatten( void )
{
	// make sure there is a terrain
	assert( m_pMap != NULL );

	// if flattening is one, then nothing would be changed, so just skip the
	// process altogether.
	if( m_uFlattening > 1 )
	{
		for( int x = 0; x < m_uSize; ++x )
		{
			for( int z = 0; z < m_uSize; ++z )
			{
				float fFlat 	= 1.0;
				float fOriginal	= GetUnscaledCell( x, z );
				
				// flatten as many times as desired
				for( int i = 0; i < m_uFlattening; ++i )
				{
					fFlat *= fOriginal;
				}
				
				// put it back into the cell
				SetCell( x, z, fFlat );
			}
		}
	}
}


//	RandomRange
//	----------------------------------------------------------------------------
float CHillTerrain::RandomRange( float min, float max )
{
	return( (((double) rand())/((double) RAND_MAX)*((double) (max - min)))+((double) min) );
}


//	license agreement:
//
//	----------------------------------------------------------------------------
//	copyright (c) 2001 bob nystrom
//
//	permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "software"), to
//	deal in the software without restriction, including without limitation the
//	rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
//	sell copies of the software, and to permit persons to whom the software is
//	furnished to do so, subject to the following conditions:
//
//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.
//
//	the software is provided "as is", without warranty of any kind, express or
//	implied, including but not limited to the warranties of merchantability,
//	fitness for a particular purpose and noninfringement. in no event shall the
//	authors or copyright holders be liable for any claim, damages or other
//	liability, whether in an action of contract, tort or otherwise, arising
//	from, out of or in connection with the software or the use or other dealings
//	in the software.
//	----------------------------------------------------------------------------
//	see www.opensource.org for more info.